﻿using System;
using System.Linq;
using System.Threading;
using HeuristicLab.Algorithms.OffspringSelectionGeneticAlgorithm;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Random;
using HeuristicLab.Selection;

namespace HeuristicLab.Problems.GaussianProcessTuning {
  public class Provider {
    public static void Main(string[] args) {
      // use random data for testing (3000 rows, 500 variables)

      int rows = 13;
      double[,] inputData = GenerateRandomData(rows, 5);
      double[] targetData = GenerateRandomData(rows);

      var prov = new Provider();
      var p = new Parameters();
      p.Generations = 1;
      p.PopulationSize = 10;
      var sol = prov.Train(targetData, inputData, null, p);
      // Console.WriteLine(sol.TestRSquared);

      //double[,] inputDataTest = GenerateRandomData(rows, 10);
      //double[] predictionTest = prov.Predict(inputDataTest, sol);
      //double[] targetDataTest = GenerateRandomData(rows);
      //OnlineCalculatorError error;
      //double r2 = OnlinePearsonsRSquaredCalculator.Calculate(predictionTest, targetDataTest, out error);
      //if (error != OnlineCalculatorError.None) Console.WriteLine(error);
      //else Console.WriteLine(r2);
    }

    // train a symb reg model
    public IRegressionSolution Train(double[] target, double[,] input, Solution oldSolution, Parameters parameters) {
      var targetVariable = "y";

      if (target.Length != input.GetLength(0)) throw new ArgumentException("length of input vectors does not match.");
      int rows = target.Length;
      int columns = input.GetLength(1);

      var variableNames =
        new string[] { targetVariable }.Concat(Enumerable.Range(1, input.GetLength(1)).Select(i => string.Format("x{0}", i)));

      double[,] combinedMatrix = new double[rows, columns + 1];
      for (int r = 0; r < rows; r++) {
        combinedMatrix[r, 0] = target[r];
        for (int c = 0; c < columns; c++) {
          combinedMatrix[r, c + 1] = input[r, c];
        }
      }

      var allowedInputVariables = variableNames.Skip(1);
      var ds = new Dataset(variableNames, combinedMatrix);
      var probData = new RegressionProblemData(ds, allowedInputVariables, targetVariable);
      probData.TrainingPartition.Start = 0;
      probData.TrainingPartition.End = rows;
      probData.TestPartition.Start = rows;
      probData.TestPartition.End = rows;
      var prob = new Problem();
      prob.ProblemDataParameter.Value = probData;
      prob.DimensionParameter.Value.Value = allowedInputVariables.Count();


      var configuredAlgorithm = CreateGeneticProgrammingAlgorithm(prob, parameters);
      // save the configured algorithm
      // XmlGenerator.Serialize(configuredAlgorithm, "osgp_" + targetVariable + ".hl");


      var resetEvent = new AutoResetEvent(false);
      configuredAlgorithm.ExceptionOccurred += (sender, args) => {
        Console.WriteLine("Exception: " + args.Value);
        configuredAlgorithm.Stop();
      };

      Solution solution = null;
      configuredAlgorithm.Stopped += (sender, args) => {
        try {
          Console.WriteLine("Run finished (execution time: {0})", configuredAlgorithm.ExecutionTime);
          solution = (Solution)configuredAlgorithm.Results["Best solution"].Value;
        }
        finally {
          resetEvent.Set();
        }
      };


      configuredAlgorithm.Prepare();
      configuredAlgorithm.Start(); // start run asynchronously
      resetEvent.WaitOne(); // wait until stopped

      //save ensemble solution
      //XmlGenerator.Serialize(ensemble, "osgp_ensemble" + targetVariable + ".hl");

      return solution;
    }

    // predict a single row
    public double Predict(double[] input, IRegressionSolution solution) {
      var allowedInputVariables = Enumerable.Range(1, input.GetLength(0)).Select(i => string.Format("x{0}", i));
      int columns = input.GetLength(0);
      double[,] combinedMatrix = new double[1, columns + 1];
      combinedMatrix[0, 0] = 0.0;
      for (int c = 0; c < columns; c++) {
        combinedMatrix[0, c + 1] = input[c];
      }

      var ds = new Dataset(allowedInputVariables.Concat(new string[] { "y" }), combinedMatrix);
      var newProbData = new RegressionProblemData(ds, allowedInputVariables, "y");
      var newSolution = solution.Model.CreateRegressionSolution(newProbData);
      return newSolution.EstimatedValues.Single();
    }


    // predict multiple rows
    public double[] Predict(double[,] input, IRegressionSolution solution) {
      var allowedInputVariables = Enumerable.Range(1, input.GetLength(1)).Select(i => string.Format("x{0}", i));
      int rows = input.GetLength(0);
      int columns = input.GetLength(1);
      double[,] combinedMatrix = new double[rows, columns + 1];
      for (int r = 0; r < rows; r++) {
        combinedMatrix[r, 0] = 0.0;
        for (int c = 0; c < columns; c++) {
          combinedMatrix[r, c + 1] = input[r, c];
        }
      }

      var ds = new Dataset(allowedInputVariables.Concat(new string[] { "y" }), combinedMatrix);
      var newProbData = new RegressionProblemData(ds, allowedInputVariables, "y");
      var newSolution = solution.Model.CreateRegressionSolution(newProbData);
      return newSolution.EstimatedValues.ToArray();
    }

    #region helper for algorithm and problem configuration
    public HeuristicOptimizationEngineAlgorithm CreateGeneticProgrammingAlgorithm(Problem problem, Parameters parameters) {
      // offspring selection GA proved to be very successful for GP
      var osga = new OffspringSelectionGeneticAlgorithm();

      problem.MaxGaussianProcessConfigurationLengthParameter.Value.Value = parameters.MaxTreeLength;
      problem.MaxGaussianProcessConfigurationDepthParameter.Value.Value = parameters.MaxTreeDepth;

      var grammar = new Grammar();

      problem.GrammarParameter.Value = grammar;

      // prepare grammar for time series prognosis 
      // combine algorithm + problem
      osga.Problem = problem;

      // set algorithm parameters
      ConfigureOffspringSelectionGeneticAlgorithmParameters<GenderSpecificSelector, SubtreeCrossover, MultiSymbolicExpressionTreeManipulator>
        (osga, parameters.PopulationSize, 0, parameters.Generations, parameters.MaxSelectionPressure, parameters.MutationRate);

      return osga;
    }

    private void ConfigureOffspringSelectionGeneticAlgorithmParameters<S, C, M>(OffspringSelectionGeneticAlgorithm osga, int popSize, int elites, int maxGens, double maxSelectionPressure, double mutationRate = 0.15, double comparisonFactorLowerBound = 1.0, double comparisonFactorUpperBound = 1.0, double successRatio = 1.0)
      where S : ISelector
      where C : ICrossover
      where M : IManipulator {
      osga.Elites.Value = elites;
      osga.MaximumGenerations.Value = maxGens;
      osga.MutationProbability.Value = mutationRate;
      osga.PopulationSize.Value = popSize;
      osga.ComparisonFactorLowerBound.Value = comparisonFactorLowerBound;
      osga.ComparisonFactorUpperBound.Value = comparisonFactorUpperBound;
      osga.SuccessRatio.Value = successRatio;
      osga.OffspringSelectionBeforeMutation.Value = false;
      osga.MaximumSelectionPressure.Value = maxSelectionPressure;

      osga.Seed.Value = 0;
      osga.SetSeedRandomly.Value = true;
      osga.Selector = osga.SelectorParameter.ValidValues
        .OfType<S>()
        .Single();

      osga.Crossover = osga.CrossoverParameter.ValidValues
        .OfType<C>()
        .Single();

      osga.Mutator = osga.MutatorParameter.ValidValues
        .OfType<M>()
        .Single();

      osga.Engine = new SequentialEngine.SequentialEngine();
    }
    #endregion


    #region helper for testing
    private static double[,] GenerateRandomData(int rows, int columns) {
      var data = new double[rows, columns];
      // generate data using a PRNG with N(0, 1) distribution
      var normalPRNG = new NormalDistributedRandom(new MersenneTwister(), 0.0, 1.0);
      for (int r = 0; r < rows; r++)
        for (int c = 0; c < columns; c++) {
          data[r, c] = normalPRNG.NextDouble();
        }
      return data;
    }
    private static double[] GenerateRandomData(int rows) {
      var data = new double[rows];
      // generate data using a PRNG with N(0, 1) distribution
      var normalPRNG = new NormalDistributedRandom(new MersenneTwister(), 0.0, 1.0);
      for (int r = 0; r < rows; r++)

        data[r] = normalPRNG.NextDouble();

      return data;
    }
    #endregion
  }
}
